#'Compute the Standardised Precipitation-Evapotranspiration Index
#' 
#'Calculation of the Standardised Precipitation-Evapotranspiration Index (SPEI) 
#'that is a multiscalar drought index based on climatic data. It can be used for 
#'determining the onset, duration and magnitude of drought conditions with 
#'respect to normal conditions. The idea behind the SPEI is to compare the 
#'highest possible evapotranspiration with the current water availability. The 
#'SPEI uses for a specific time frequency the difference between precipitation 
#'and potential evapotranspiration. This represents a simple climatic water 
#'balance which is calculated at different time scales to obtain the SPEI. This 
#'function is build to be compatible with other tools in that work with 
#''s2dv_cube' object class. The input data must be this object class. If you 
#'don't work with 's2dv_cube', see PeriodSPEI.
#' 
#'Next, some specifications for the calculation of this indicator will be 
#'discussed. On the one hand, the model to be used to calculate Potential 
#'Evapotranspiration is specified with the 'pet_method' parameter (it can be 
#'hargreaves, hargraves modified or thornwhite). On the other hand, to choose 
#'the time scale in which you want to accumulate the SPEI (SPEI3, SPEI6...) is 
#'done using the 'accum' parameter, where you must indicate the number of time 
#'steps you want to accumulate throughout 'leadtime_dim'. The accumulation will 
#'be performed backwards by default. Since the accumulation is done for the 
#'elapsed time steps, there will be no complete accumulations until reaching the 
#'time instant equal to the value of the 'accum' parameter. For this reason, in 
#'the result, we will find that for the dimension where the accumulation has 
#'been carried out, the values of the array will be NA since they do not include 
#'complete accumulations. Also, there is a parameter to specify if the 
#'standardization that can be TRUE or FALSE. If it is TRUE, the data is fit to a 
#'probability distribution to transform the original values to standardized 
#'units that are comparable in space and time and at different SPEI time scales. 
#'The 'na.rm' parameter is a logical parameter used to decide whether to remove 
#'the NA values from the data before doing the calculation. It must be taken 
#'into account that if 'na.rm' is FALSE and there is some NA value in the 
#'specific coordinates which the SPEI is computed, standardization cannot be 
#'carried out for those coordinates and therefore, the result will be filled 
#'with NA for the specific coordinates. However, when 'na.rm' is TRUE,  if the 
#'amount of data for those specific coordinates is smaller than 4, it will not 
#'be possible to carry out because we will not have enough data and the result 
#'will be also filled with NAs for that coordinates. When only 'exp' is provided 
#'('exp_cor' is NULL) the Standardization is computed with cross validation. For 
#'more information about SPEI calculation, see functions CST_PeriodPET, CST_PeriodAccumulation 
#'and CST_PeriodStandardization. 
#' 
#'@param exp A named list with the needed \code{s2dv_cube} objects containing 
#'  the seasonal forecast experiment in the data element for each variable. 
#'  Specific variables are needed for each method used in computing the 
#'  Potential Evapotranspiration. See parameter 'pet_method'. The accepted 
#'  variables are: 'tmin' and 'tmax', required for methods 'hargreaves' and 
#'  'hargreaves_modified' and 'tmean', required for method 'thornthwaite'. 
#'  Variable 'pr' is always needed. The units for temperature variables
#'  ('tmin', 'tmax' and 'tmean') need to be in Celcius degrees; the units for 
#'  precipitation ('pr') need to be in mm/month. 
#'@param exp_cor A named list with the needed \code{s2dv_cube} objects for each 
#'  variable in which the quantile PeriodSPEI should be applied. If it is not 
#'  specified, the PeriodSPEI is calculated from object 'exp'.
#'@param time_dim A character string indicating the name of the temporal 
#'  dimension. By default, it is set to 'syear'. 
#'@param leadtime_dim A character string indicating the name of the temporal 
#'  dimension. By default, it is set to 'time'. 
#'@param memb_dim A character string indicating the name of the dimension in 
#'  which the ensemble members are stored. When set it to NULL, threshold is 
#'  computed for individual members.
#'@param lat_dim A character string indicating the name of the latitudinal 
#'  dimension. By default it is set by 'latitude'.
#'@param accum An integer value indicating the number of months for the 
#'  accumulation for each variable. When it is greater than 1, the result will 
#'  be filled with NA until the accum 'time_dim' dimension number due to the 
#'  accumulation to previous months. The accumulation is performed backwards 
#'  by default.
#'@param ref_period A list with two numeric values with the starting and end 
#'  points of the reference period used for computing the index. The default 
#'  value is NULL indicating that the first and end values in data will be 
#'  used as starting and end points.
#'@param params A multidimensional array with named dimensions for computing the 
#'  SPEI. This option overrides computation of fitting parameters. It needs 
#'  to have the following dimensions: same leadtime dimension of 'exp' 
#'  (specified in 'leadtime_dim'); time dimension of length 1 (specified in 
#'  'time_dim'); and a dimension named 'coef' with the length of the 
#'  coefficients needed for the used distribution (for 'Gamma' coef dimension is 
#'  of lenght 2, for 'log-Logistic' is of length 3). It can't have member 
#'  dimension (specified in 'memb_dim').
#'@param pet_exp A multidimensional array containing the Potential  
#'  EvapoTranspiration data of 'exp'. It must have the same dimensions of the 
#'  variable 'pr' of 'exp'. If it is NULL it is calculated using the provided 
#'  variables with specified 'pet_method'. It is NULL by default.
#'@param pet_expcor A multidimensional array containing the Potential  
#'  EvapoTranspiration data of 'exp_cor'. It must have the same dimensions of 
#'  the variable 'pr' of 'exp_cor'. If it is NULL it is calculated using the 
#'  provided variables with specified 'pet_method'. It is NULL by default.
#'@param standardization A logical value indicating wether the standardization 
#'  is computed.
#'@param pet_method A character string indicating the method used to compute 
#'  the potential evapotranspiration. The accepted methods are:
#'  'hargreaves' and 'hargreaves_modified', that require the data to have 
#'  variables tmin and tmax; and 'thornthwaite', that requires variable 
#'  'tmean'.  
#'@param method A character string indicating the standardization method used. 
#'  If can be: 'parametric' or 'non-parametric'. It is set to 'parametric' by 
#'  default.
#'@param distribution A character string indicating the name of the distribution 
#'  function to be used for computing the SPEI. The accepted names are: 
#'  'log-Logistic' and 'Gamma'. It is set to 'log-Logistic' by default. The 
#'  'Gamma' method only works when only precipitation is provided and other 
#'  variables are 0 because it is positive defined (SPI indicator). If other  
#'  distribution functions want to be used contact the authors.
#'@param handle_infinity A logical value wether to return Infinite values (TRUE)
#'  or not (FALSE).
#'@param na.rm A logical value indicating whether NA values should be removed 
#'  from data. It is FALSE by default. If it is FALSE and there are NA values, 
#'  (if standardization is TRUE) all values of other dimensions except time_dim 
#'  and leadtime_dim will be set to NA directly. On the other hand, if it is 
#'  TRUE, if the data from other dimensions except time_dim and leadtime_dim is 
#'  not reaching 4 values, it is not enough values to estimate the parameters 
#'  and the result will include NA.
#'@param ncores An integer value indicating the number of cores to use in 
#'  parallel computation.
#' 
#'@return An 's2dv_cube' object containing the SPEI multidimensional array in 
#'element \code{data} with same dimensions as input data. If 'exp_cor' is 
#'provided, only results from 'exp_cor' will be provided. The parameters of the 
#'standardization will only be returned if 'return_params' is TRUE. The SPEI 
#'will only be computed if 'standardization' is TRUE. If 'standardization' is 
#'FALSE, only the climatic water balance (precipitation minus 
#'evapotranspiration) will be returned. The resultant arrays will have the same 
#'dimensions as the initial input data. The other elements in the 's2dv_cube' 
#'will be updated with the combined information of the input data arrays.
#' 
#'@examples 
#'dims <- c(var = 1, sday = 1, sweek = 1, syear = 6, time = 3, 
#'          latitude = 2, longitude = 1, ensemble = 25)
#'dimscor <- c(var = 1, sday = 1, sweek = 1, syear = 1, time = 3, 
#'             latitude = 2, longitude = 1, ensemble = 15)
#'
#'dates_exp <- as.POSIXct(c(paste0(2010:2015, "-08-16"), 
#'                          paste0(2010:2015, "-09-15"), 
#'                          paste0(2010:2015, "-10-16")), 'UTC')
#'dim(dates_exp) <- c(sday = 1, sweek = 1, syear = 6, time = 3)
#'dates_expcor <- as.POSIXct(c(paste0(2020, "-08-16"), paste0(2020, "-09-15"), 
#'                             paste0(2020, "-10-16")), 'UTC')
#'dim(dates_expcor) <- c(sday = 1, sweek = 1, syear = 1, time = 3)
#'lat <- c(40,40.1)
#'
#'exp_tmax <- array(rnorm(900, 27.73, 5.26), dim = dims)
#'exp_tmin <- array(rnorm(900, 14.83, 3.86), dim = dims)
#'exp_pr <- array(rnorm(900, 21.19, 25.64), dim = dims)
#'
#'expcor_tmax <- array(rnorm(540, 29.03, 5.67), dim = dimscor)
#'expcor_tmin <- array(rnorm(540, 15.70, 4.40), dim = dimscor)
#'expcor_pr <- array(rnorm(540, 15.62, 21.38), dim = dimscor)
#'
#'exp <- list('tmax' = exp_tmax, 'tmin' = exp_tmin, 'pr' = exp_pr)
#'exp_cor <- list('tmax' = expcor_tmax, 'tmin' = expcor_tmin, 
#'                'pr' = expcor_pr)
#'
#'exp <- lapply(exp, CSTools::s2dv_cube, coords = list(latitude = lat), 
#'              Dates = dates_exp)
#'exp_cor <- lapply(exp_cor, CSTools::s2dv_cube, coords = list(latitude = lat), 
#'                  Dates = dates_expcor)
#'
#'res <- CST_PeriodSPEI(exp = exp, exp_cor = exp_cor)
#' 
#'@importFrom CSTools s2dv_cube
#'@export
CST_PeriodSPEI <- function(exp, exp_cor = NULL, time_dim = 'syear', 
                           leadtime_dim = 'time', memb_dim =  'ensemble', 
                           lat_dim = 'latitude', accum = 1, ref_period = NULL, 
                           params = NULL, pet_exp = NULL, pet_expcor = NULL, 
                           standardization = TRUE, pet_method = 'hargreaves', 
                           method = 'parametric', distribution = 'log-Logistic',
                           handle_infinity = FALSE, return_params = FALSE, 
                           na.rm = FALSE, ncores = NULL) {
                          
  # Check 's2dv_cube'
  if (is.null(exp)) {
    stop("Parameter 'exp' cannot be NULL.")
  }
  if (!all(sapply(exp, function(x) inherits(x, 's2dv_cube')))) {
    stop("Parameter 'exp' must be a list of 's2dv_cube' class.")
  }
  if (!is.null(exp_cor)) {
    if (!all(sapply(exp_cor, function(x) inherits(x, 's2dv_cube')))) {
      stop("Parameter 'exp_cor' must be a list of 's2dv_cube' class.")
    }
  }
  # latitude
  if (!any(names(exp[[1]]$coords) %in% .KnownLatNames())) {
    stop("Spatial coordinate names of parameter 'exp' do not match any ",
         "of the names accepted by the package.")
  }
  # Dates
  dates_exp <- exp[[1]]$attrs$Dates
  if (!'Dates' %in% names(exp[[1]]$attrs)) {
    stop("Element 'Dates' is not found in 'attrs' list of 'exp'. ", 
         "See 's2dv_cube' object description in README file for more ", 
         "information.")
  }

  if (!is.null(exp_cor)) {
    if (!'Dates' %in% names(exp_cor[[1]]$attrs)) {
      stop("Element 'Dates' is not found in 'attrs' list of 'exp_cor'. ", 
           "See 's2dv_cube' object description in README file for more ", 
           "information.")
    }
  }

  lat_dim <- names(exp[[1]]$coords)[[which(names(exp[[1]]$coords) %in% .KnownLatNames())]]

  res <- PeriodSPEI(exp = lapply(exp, function(x) x$data), 
                    dates_exp = exp[[1]]$attrs$Dates, 
                    lat = exp[[1]]$coords[[lat_dim]], 
                    exp_cor = if (is.null(exp_cor)) {NULL} else {
                                  lapply(exp_cor, function(x) x$data)}, 
                    dates_expcor = exp_cor[[1]]$attrs$Dates, 
                    time_dim = time_dim, leadtime_dim = leadtime_dim, 
                    memb_dim = memb_dim, lat_dim = lat_dim, 
                    accum = accum, ref_period = ref_period, params = params, 
                    pet_exp = pet_exp, pet_expcor = pet_expcor, 
                    standardization = standardization, 
                    pet_method = pet_method, method = method, 
                    distribution = distribution, 
                    handle_infinity = handle_infinity, 
                    return_params = return_params, na.rm = na.rm, 
                    ncores = ncores)

  if (!is.null(exp_cor)) {
    source_files_expcor <- lapply(exp_cor, function(x) {x$attrs$source_files})
    source_files <- lapply(exp, function(x) {x$attrs$source_files})
    source_files <- c(exp = source_files, exp_cor = source_files_expcor)
    coords <- exp_cor[[1]]$coords
    Dates <- exp_cor[[1]]$attrs$Dates
    metadata <- exp_cor[[1]]$attrs$Variable$metadata
    metadata_names <- names(metadata)
  } else {
    source_files <- lapply(exp, function(x) {x$attrs$source_files})
    coords <- exp[[1]]$coords
    Dates <- exp[[1]]$attrs$Dates
    metadata <- exp[[1]]$attrs$Variable$metadata
    metadata_names <- names(metadata) 
  }

  if (standardization) {
    varname <- 'SPEI'
  } else {
    varname <- 'Precipitation minus accumulated PET'
  }

  if (return_params & standardization) {
    metadata_names <- intersect(names(dim(res[[1]])), metadata_names)
    suppressWarnings(
      res[[1]] <- CSTools::s2dv_cube(data = res[[1]], coords = coords, 
                                     varName = varname, 
                                     metadata = metadata[metadata_names],
                                     Dates = Dates, 
                                     source_files = source_files,
                                     when = Sys.time())
    )
    return(list(spei = res[[1]], params = res[[2]]))
  } else {
    metadata_names <- intersect(names(dim(res)), metadata_names)
    suppressWarnings(
      res <- CSTools::s2dv_cube(data = res, coords = coords, 
                                varName = varname, 
                                metadata = metadata[metadata_names],
                                Dates = Dates, 
                                source_files = source_files, 
                                when = Sys.time())
    )
    return(res)
  }
}


#'Compute the Standardised Precipitation-Evapotranspiration Index
#' 
#'Calculation of the Standardised Precipitation-Evapotranspiration Index (SPEI) 
#'that is a multiscalar drought index based on climatic data. It can be used for 
#'determining the onset, duration and magnitude of drought conditions with 
#'respect to normal conditions. The idea behind the SPEI is to compare the 
#'highest possible evapotranspiration with the current water availability. The 
#'SPEI uses for a specific time frequency the difference between precipitation 
#'and potential evapotranspiration. 
#' 
#'Next, some specifications for the calculation of this indicator will be 
#'discussed. On the one hand, the model to be used to calculate Potential 
#'Evapotranspiration is specified with the 'pet_method' parameter (it can be 
#'hargreaves, hargraves modified or thornwhite). On the other hand, to choose 
#'the time scale in which you want to accumulate the SPEI (SPEI3, SPEI6...) is 
#'done using the 'accum' parameter, where you must indicate the number of time 
#'steps you want to accumulate throughout 'leadtime_dim'. The accumulation will 
#'be performed backwards by default. Since the accumulation is done for the 
#'elapsed time steps, there will be no complete accumulations until reaching the 
#'time instant equal to the value of the 'accum' parameter. For this reason, in 
#'the result, we will find that for the dimension where the accumulation has 
#'been carried out, the values of the array will be NA since they do not include 
#'complete accumulations. Also, there is a parameter to specify if the 
#'standardization that can be TRUE or FALSE. If it is TRUE, the data is fit to a 
#'probability distribution to transform the original values to standardized 
#'units that are comparable in space and time and at different SPEI time scales. 
#'The 'na.rm' parameter is a logical parameter used to decide whether to remove 
#'the NA values from the data before doing the calculation. It must be taken 
#'into account that if 'na.rm' is FALSE and there is some NA value in the 
#'specific coordinates which the SPEI is computed, standardization cannot be 
#'carried out for those coordinates and therefore, the result will be filled 
#'with NA for the specific coordinates. However, when 'na.rm' is TRUE,  if the 
#'amount of data for those specific coordinates is smaller than 4, it will not 
#'be possible to carry out because we will not have enough data and the result 
#'will be also filled with NAs for that coordinates. When only 'exp' is provided 
#'('exp_cor' is NULL) the Standardization is computed with cross validation. For 
#'more information about SPEI calculation, see functions PeriodPET, 
#'PeriodAccumulation and PeriodStandardization. 
#' 
#'@param exp A named list with multidimensional array objects containing 
#'  the seasonal forecast experiment in the data element for each variable. 
#'  Specific variables are needed for each method used in computing the 
#'  Potential Evapotranspiration. See parameter 'pet_method'. The accepted 
#'  variables are: 'tmin' and 'tmax', required for methods 'hargreaves' and 
#'  'hargreaves_modified' and 'tmean', required for method 'thornthwaite'. 
#'  Variable 'pr' is always needed. The units for temperature variables
#'  ('tmin', 'tmax' and 'tmean') need to be in Celcius degrees; the units for 
#'  precipitation ('pr') need to be in mm/month. 
#'@param dates_exp An array of temporal dimensions containing the Dates of 
#'  'exp'. It must be of class 'Date' or 'POSIXct'.
#'@param lat A numeric vector containing the latitude values of 'exp'.
#'@param exp_cor A named list with multidimensional array objects for each 
#'  variable in which the quantile PeriodSPEI should be applied. If it is not 
#'  specified, the PeriodSPEI is calculated from object 'exp'.
#'@param dates_expcor An array of temporal dimensions containing the Dates of 
#'  'exp_cor'. It must be of class 'Date' or 'POSIXct'.
#'@param time_dim A character string indicating the name of the temporal 
#'  dimension. By default, it is set to 'syear'. 
#'@param leadtime_dim A character string indicating the name of the temporal 
#'  dimension. By default, it is set to 'time'. 
#'@param memb_dim A character string indicating the name of the dimension in 
#'  which the ensemble members are stored. When set it to NULL, threshold is 
#'  computed for individual members.
#'@param lat_dim A character string indicating the name of the latitudinal 
#'  dimension. By default it is set by 'latitude'.
#'@param accum accum An integer value indicating the number of months for the 
#'  accumulation for each variable. When it is greater than 1, the result will 
#'  be filled with NA until the accum time_dim dimension number due to the 
#'  accumulation to previous months. The accumulation is performed backwards 
#'  by default.
#'@param ref_period A list with two numeric values with the starting and end 
#'  points of the reference period used for computing the index. The default 
#'  value is NULL indicating that the first and end values in data will be 
#'  used as starting and end points.
#'@param params A multidimensional array with named dimensions for computing the 
#'  SPEI. This option overrides computation of fitting parameters. It needs 
#'  to be of same time dimension (specified in 'time_dim') of 'exp' and a 
#'  dimension named 'coef' with the length of the coefficients needed for the 
#'  used distribution (for 'Gamma' coef dimension is of lenght 2, for 
#'  'log-Logistic' is of length 3). It also needs to have a leadtime dimension 
#'  (specified in 'leadtime_dim') of length 1.
#'@param pet_exp A multidimensional array containing the Potential  
#'  EvapoTranspiration data of 'exp'. It must have the same dimensions of the 
#'  variable 'pr' of 'exp'. If it is NULL it is calculated using the provided 
#'  variables with specified 'pet_method'. It is NULL by default.
#'@param pet_expcor A multidimensional array containing the Potential  
#'  EvapoTranspiration data of 'exp_cor'. It must have the same dimensions of 
#'  the variable 'pr' of 'exp_cor'. If it is NULL it is calculated using the 
#'  provided variables with specified 'pet_method'. It is NULL by default.
#'@param standardization A logical value indicating wether the standardization 
#'  is computed.
#'@param pet_method A character string indicating the method used to compute 
#'  the potential evapotranspiration. The accepted methods are:
#'  'hargreaves' and 'hargreaves_modified', that require the data to have 
#'  variables tmin and tmax; and 'thornthwaite', that requires variable 
#'  'tmean'.  
#'@param method A character string indicating the standardization method used. 
#'  If can be: 'parametric' or 'non-parametric'.
#'@param distribution A character string indicating the name of the distribution 
#'  function to be used for computing the SPEI. The accepted names are: 
#'  'log-Logistic' and 'Gamma'. It is set to 'log-Logistic' by default. The 
#'  'Gamma' method only works when only precipitation is provided and other 
#'  variables are 0 because it is positive defined (SPI indicator).
#'@param handle_infinity A logical value wether to return Infinite values (TRUE)
#'  or not (FALSE).
#'@param return_params A logical value indicating wether to return parameters 
#'  array (TRUE) or not (FALSE). It is FALSE by default.
#'@param na.rm A logical value indicating whether NA values should be removed 
#'  from data. It is FALSE by default. If it is FALSE and there are NA values, 
#'  (if standardization is TRUE) all values of other dimensions except time_dim 
#'  and leadtime_dim will be set to NA directly. On the other hand, if it is 
#'  TRUE, if the data from other dimensions except time_dim and leadtime_dim is 
#'  not reaching 4 values, it is not enough values to estimate the parameters 
#'  and the result will include NA.
#'@param ncores An integer value indicating the number of cores to use in 
#'  parallel computation.
#' 
#'@return A multidimensional array containing the SPEI with same dimensions as 
#'input data. If 'exp_cor' is provided, only results from 'exp_cor' will be 
#'provided. The parameters of the standardization will only be returned 
#'if 'return_params' is TRUE. The SPEI will only be computed if 
#''standardization' is TRUE. If 'standardization' is FALSE, only the climatic 
#'water balance (precipitation minus evapotranspiration) will be returned. The 
#'resultant arrays will have the same dimensions as the initial input data.
#' 
#'@examples 
#'dims <- c(var = 1, sday = 1, sweek = 1, syear = 6, time = 3, 
#'          latitude = 2, longitude = 1, ensemble = 25)
#'dimscor <- c(var = 1, sday = 1, sweek = 1, syear = 1, time = 3, 
#'             latitude = 2, longitude = 1, ensemble = 15)
#'exp_tmax <- array(rnorm(900, 27.73, 5.26), dim = dims)
#'exp_tmin <- array(rnorm(900, 14.83, 3.86), dim = dims)
#'exp_pr <- array(rnorm(900, 21.19, 25.64), dim = dims)
#' 
#'expcor_tmax <- array(rnorm(540, 29.03, 5.67), dim = dimscor)
#'expcor_tmin <- array(rnorm(540, 15.70, 4.40), dim = dimscor)
#'expcor_pr <- array(rnorm(540, 15.62, 21.38), dim = dimscor)
#' 
#'dates_exp <- as.Date(c(paste0(2010:2015, "-08-16"), 
#'                       paste0(2010:2015, "-09-15"), 
#'                       paste0(2010:2015, "-10-16")))
#'dim(dates_exp) <- c(sday = 1, sweek = 1, syear = 6, time = 3)
#'dates_expcor <- as.POSIXct(c(paste0(2020, "-08-16"), 
#'                             paste0(2020, "-09-15"), 
#'                             paste0(2020, "-10-16")), 'UTC')
#'dim(dates_expcor) <- c(sday = 1, sweek = 1, syear = 1, time = 3)
#'lat <- c(40,40.1)
#' 
#'exp <- list('tmax' = exp_tmax, 'tmin' = exp_tmin, 'pr' = exp_pr)
#'exp_cor <- list('tmax' = expcor_tmax, 'tmin' = expcor_tmin, 
#'                'pr' = expcor_pr)
#'res <- PeriodSPEI(exp = exp, exp_cor = exp_cor, lat = lat, 
#'                  dates_exp = dates_exp, dates_expcor = dates_expcor)
#' 
#'@export
PeriodSPEI <- function(exp, dates_exp, lat, exp_cor = NULL, dates_expcor = NULL,
                       time_dim = 'syear', leadtime_dim = 'time', 
                       memb_dim = 'ensemble', lat_dim = 'latitude',
                       accum = 1, ref_period = NULL, params = NULL, 
                       pet_exp = NULL, pet_expcor = NULL, standardization = TRUE, 
                       pet_method = 'hargreaves', method = 'parametric', 
                       distribution = 'log-Logistic', handle_infinity = FALSE, 
                       return_params = FALSE, na.rm = FALSE, ncores = NULL) {

  # Initial checks
  ## exp
  if (!inherits(exp, 'list')) {
    stop("Parameter 'exp' needs to be a named list with the needed variables.")
  }
  if (is.null(names(exp))) {
    stop("Parameter 'exp' needs to be a named list with the variable names.")
  }
  if (any(sapply(exp, function(x) is.null(names(dim(x)))))) {
    stop("Parameter 'exp' needs to be a list of arrays with dimension names.")
  }
  dims <- lapply(exp, function(x) dim(x))
  first_dims <- dims[[1]]
  all_equal <- all(sapply(dims[-1], function(x) identical(first_dims, x)))
  if (!all_equal) {
    stop("Parameter 'exp' variables need to have the same dimensions.")
  }

  ## exp_cor
  if (!is.null(exp_cor)) {
    if (!inherits(exp_cor, 'list')) {
      stop("Parameter 'exp_cor' needs to be a named list with the needed ", 
           "variables if it is not NULL.")
    }
    if (is.null(names(exp_cor))) {
      stop("Parameter 'exp_cor' needs to be a named list with the variable names.")
    }
    if (any(sapply(exp_cor, function(x) is.null(names(dim(x)))))) {
      stop("Parameter 'exp_cor' needs to be a list of arrays with dimension names.")
    }
    dimscor <- lapply(exp_cor, function(x) dim(x))
    first_dimscor <- dimscor[[1]]
    all_equal <- all(sapply(dimscor[-1], function(x) identical(first_dimscor, x)))
    if (!all_equal) {
      stop("Parameter 'exp_cor' variables need to have the same dimensions.")
    }
  }
  # Variable checks
  ## exp (2)
  pet <- vector("list", 2)
  if (!('pr' %in% names(exp))) {
    stop("Variable 'pr' is not included in 'exp'.")
  }
  ## exp_cor (2)
  if (!is.null(exp_cor)) {
    if (!('pr' %in% names(exp_cor))) {
      stop("Variable 'pr' is not included in 'exp_cor'.")
    }
    if (length(pet_method) == 1) {
      pet_method <- rep(pet_method, 2)
    }
  }
  ## time_dim
  if (!is.character(time_dim) | length(time_dim) != 1) {
    stop("Parameter 'time_dim' must be a character string.")
  }
  if (!all(sapply(exp, function(x) time_dim %in% names(dim(x))))) {
    stop("Parameter 'time_dim' is not found in 'exp' dimension.")
  }
  if (!is.null(exp_cor)) {
    if (!all(sapply(exp_cor, function(x) time_dim %in% names(dim(x))))) {
      stop("Parameter 'time_dim' is not found in 'exp_cor' dimension.")
    }
  }
  ## leadtime_dim
  if (!is.character(leadtime_dim) | length(leadtime_dim) != 1) {
    stop("Parameter 'leadtime_dim' must be a character string.")
  }
  if (!all(sapply(exp, function(x) leadtime_dim %in% names(dim(x))))) {
    stop("Parameter 'leadtime_dim' is not found in 'exp' dimension.")
  }
  if (!is.null(exp_cor)) {
    if (!all(sapply(exp_cor, function(x) leadtime_dim %in% names(dim(x))))) {
      stop("Parameter 'leadtime_dim' is not found in 'exp_cor' dimension.")
    }
  }
  ## memb_dim
  if (!is.character(memb_dim) | length(memb_dim) != 1) {
    stop("Parameter 'memb_dim' must be a character string.")
  }
  if (!all(sapply(exp, function(x) memb_dim %in% names(dim(x))))) {
    stop("Parameter 'memb_dim' is not found in 'exp' dimension.")
  }
  if (!is.null(exp_cor)) {
    if (!all(sapply(exp_cor, function(x) memb_dim %in% names(dim(x))))) {
      stop("Parameter 'memb_dim' is not found in 'exp_cor' dimension.")
    }
  }
  ## lat_dim
  if (!is.character(lat_dim) | length(lat_dim) != 1) {
    stop("Parameter 'lat_dim' must be a character string.")
  }
  if (!all(sapply(exp, function(x) lat_dim %in% names(dim(x))))) {
    stop("Parameter 'lat_dim' is not found in 'exp' dimension.")
  }
  if (!is.null(exp_cor)) {
    if (!all(sapply(exp_cor, function(x) lat_dim %in% names(dim(x))))) {
      stop("Parameter 'lat_dim' is not found in 'exp_cor' dimension.")
    }
  }
  ## dates
  if (is.null(dates_exp)) {
    stop("Parameter 'dates_exp' is missing, dates must be provided.")
  }
  if (!any(inherits(dates_exp, 'Date'), inherits(dates_exp, 'POSIXct'))) {
    stop("Parameter 'dates_exp' is not of the correct class, ", 
         "only 'Date' and 'POSIXct' classes are accepted.")
  }
  if (!time_dim %in% names(dim(dates_exp)) | !leadtime_dim %in% names(dim(dates_exp))) {
    stop("Parameter 'dates_exp' must have 'time_dim' and 'leadtime_dim' ",
         "dimension.")
  }
  if (!all(dim(exp[[1]])[c(time_dim, leadtime_dim)] == 
           dim(dates_exp)[c(time_dim, leadtime_dim)])) {
    stop("Parameter 'dates_exp' needs to have the same length as 'time_dim' ", 
         "and 'leadtime_dim' as 'exp'.")
  }
  if (!is.null(exp_cor)) {
    if (is.null(dates_expcor)) {
      stop("Parameter 'dates_expcor' is missing, dates for 'exp_cor' must be ", 
           "provided if exp_cor is not NULL.")
    }
    if (!any(inherits(dates_expcor, 'Date'), inherits(dates_expcor, 'POSIXct'))) {
      stop("Element 'Dates' in 'exp_cor' is not of the correct class, ", 
           "only 'Date' and 'POSIXct' classes are accepted.")
    }
    if (!time_dim %in% names(dim(dates_expcor)) | !leadtime_dim %in% names(dim(dates_expcor))) {
      stop("Parameter 'dates_expcor' must have 'time_dim' and 'leadtime_dim' ",
           "dimension.")
    }
    if (!all(dim(exp_cor[[1]])[c(time_dim, leadtime_dim)] == 
             dim(dates_expcor)[c(time_dim, leadtime_dim)])) {
      stop("Parameter 'dates_expcor' needs to have the same length as ", 
           "'time_dim' and 'leadtime_dim' as 'exp_cor'.")
    }
  }
  ## accum
  if (accum > dim(exp[[1]])[leadtime_dim]) {
    stop(paste0("Cannot compute accumulation of ", accum, " months because ", 
                "loaded data has only ", dim(exp[[1]])[leadtime_dim], " months."))
  }
  ## standardization
  if (!is.logical(standardization)) {
    stop("Parameter 'standardization' must be a logical value.")
  }
  ## na.rm
  if (!is.logical(na.rm)) {
    stop("Parameter 'na.rm' must be logical.")
  }

  # Complete dates
  dates <- .return2list(dates_exp, dates_expcor)

  # Compute PeriodSPEI
  k = 0
  spei_res <- NULL
  computed_pet <- FALSE
  accumulated <- list(NULL, NULL)
  for (data in .return2list(exp, exp_cor)) {
    k = k + 1
    # Evapotranspiration estimation (unless pet is already provided)
    if (is.null(pet[[k]]) | computed_pet) {
      pet[[k]] <- PeriodPET(data = data, dates = dates[[k]], 
                            lat = lat, pet_method = pet_method[k], 
                            time_dim = time_dim, leadtime_dim = leadtime_dim, 
                            lat_dim = lat_dim, na.rm = na.rm, ncores = ncores)
      computed_pet <- TRUE
    }
    if (!na.rm & any(is.na(pet[[k]]))) {
      mask_na <- which(is.na(pet[[k]]))
      warning("There are NAs in PET.")
    }

    # Accumulation
    diff_p_pet <- data$pr - pet[[k]]
    accumulated[[k]] <- PeriodAccumulation(data = diff_p_pet, dates = dates[[k]],
                                           rollwidth = accum, time_dim = leadtime_dim,
                                           sdate_dim = time_dim, frequency = 'monthly', 
                                           ncores = ncores)
  }

  if (standardization) {
    # Standardization
    spei <- PeriodStandardization(data = accumulated[[1]], 
                                  data_cor = accumulated[[2]],
                                  dates = dates[[1]],
                                  params = params, 
                                  time_dim = time_dim,
                                  leadtime_dim = leadtime_dim,
                                  memb_dim = memb_dim, ref_period = ref_period,
                                  handle_infinity = handle_infinity,
                                  return_params = return_params,
                                  method = method, distribution = distribution, 
                                  na.rm = TRUE, ncores = ncores)
  } else {
    spei <- accumulated[[k]]
  }
  return(spei)
}