#'@rdname Predictability
#'@title Computing scores of predictability using two dynamical proxies 
#'based on dynamical systems theory.
#'
#'@author Carmen Alvarez-Castro, \email{carmen.alvarez-castro@cmcc.it}
#'@author Maria M. Chaves-Montero, \email{mdm.chaves-montero@cmcc.it}
#'@author Veronica Torralba, \email{veronica.torralba@cmcc.it}
#'@author Davide Faranda, \email{davide.faranda@lsce.ipsl.fr}
#'
#'@description This function divides in terciles the two dynamical proxies 
#'computed with CST_ProxiesAttractor or ProxiesAttractor. These terciles will
#'be used to measure the predictability of the system in dyn_scores. When the
#'local dimension 'dim' is small and the inverse of persistence 'theta' is 
#'small the predictability is high, and viceversa. 
#'
#'@references Faranda, D., Alvarez-Castro, M.C., Messori, G., Rodriguez, D., 
#'and Yiou, P. (2019). The hammam effect or how a warm ocean enhances large 
#'scale atmospheric predictability.Nature Communications, 10(1), 1316. 
#'DOI = https://doi.org/10.1038/s41467-019-09305-8 "
#'@references Faranda, D., Gabriele Messori and Pascal Yiou. (2017).
#'Dynamical proxies of North Atlantic predictability and extremes. 
#'Scientific Reports, 7-41278, 2017.
#'
#'@param dim A multidimensional array with named dimensions containing the local 
#'  dimension as the output of CST_ProxiesAttractor or ProxiesAttractor. 
#'@param theta A multidimensional array with named dimensions containing the  
#'  inverse of the persistence 'theta' as the output of CST_ProxiesAttractor or 
#'  ProxiesAttractor.
#'@param ncores The number of cores to use in parallel computation.
#'
#'@return A list of length 3:
#'\itemize{
#'  \item\code{pred.dim} {A list of two lists 'qdim' and 'pos.d'. The 'qdim' 
#'  list contains values of local dimension 'dim' divided by terciles: 
#'    \itemize{
#'      \item{d1: lower tercile (high predictability),}
#'      \item{d2: middle tercile,}
#'      \item{d3: higher tercile (low predictability)}
#'    }
#'  The 'pos.d' list contains the position of each tercile in parameter 'dim':}
#'  \item\code{pred.theta} {A list of two lists 'qtheta' and 'pos.t'. The 'qtheta' 
#'  list contains values of the inverse of persistence 'theta' divided by 
#'  terciles:
#'    \itemize{
#'      \item{th1: lower tercile (high predictability),}
#'      \item{th2: middle tercile,}
#'      \item{th3: higher tercile (low predictability)} 
#'    } 
#'  The 'pos.t' list contains the position of each tercile in parameter 'theta'.
#'  }
#'  \item\code{dyn_scores} {An array with named dimensions containing values 
#'  from 0 to 1. A dyn_score of 1 indicates the highest predictability.}
#'}
#'
#'@examples
#'# Creating an example of matrix dat(time,grids):
#'m <- matrix(rnorm(2000) * 10, nrow = 50, ncol = 40)
#'names(dim(m)) <- c('time', 'grid')
#'# imposing a threshold
#'quanti <-  0.90
#'# computing dyn_scores from parameters dim and theta of the attractor
#'attractor <- ProxiesAttractor(dat = m, quanti = 0.60)
#'predyn <- Predictability(dim = attractor$dim, theta = attractor$theta)
#'@export 
Predictability <- function(dim, theta, time_dims = c('ftime', 'sdate'),
                           ncores = NULL) {
  # Initial checks
  # dim and theta
  if (is.null(dim) | is.null(theta)) {
    stop("Parameters 'dim' and 'theta' cannot be NULL.")
  }
  # temporal dimensions
  if (!all(time_dims %in% names(dim(dim)))) {
    stop("Parameter 'time_dims' is not found in 'dim' dimension names.")
  }
  if (!all(time_dims %in% names(dim(theta)))) {
    stop("Parameter 'time_dims' is not found in 'theta' dimension names.")
  }
  if (!all(dim(dim)[time_dims] %in% dim(theta)[time_dims])) {
    stop("Dimensions 'temporals' of 'dim' and 'theta' must be the same.")
  }
  if (length(time_dims) > 1) {
    dim <- MergeDims(dim, merge_dims = time_dims, rename_dim = 'temporals')
    theta <- MergeDims(theta, merge_dims = time_dims, rename_dim = 'temporals')
  }

  # The number of elements in each tercile can vary and 
  # Apply requires to merge them being of the same length
  # there will be NAs in the outputs

  length_theta <- Apply(list(theta), target_dims = 'temporals',
                        function(x) {
                          th <- .terciles_position(x)
                          max(length(th[[1]]), length(th[[2]]), length(th[[3]]))},
                          ncores = ncores)$output1
  length_dim <- Apply(list(dim), target_dims = 'temporals',
                      function(x) {
                        d <- .terciles_position(x)
                        max(length(d[[1]]), length(d[[2]]), length(d[[3]]))},
                        ncores = ncores)$output1
  length_theta <- max(length_theta)
  length_dim <- max(length_dim)
  pred <- Apply(list(dim, theta), target_dims = 'temporals',
                fun = .predictability, out_len = c(length_dim, length_theta),  
                ncores = ncores)
  dim(pred$dyn_scores) <- dim(theta)
  return(list(pred.dim = list(qdim = list(d1 = pred$qdim.d1,
                                          d2 = pred$qdim.d2,
                                          d3 = pred$qdim.d3),
                              pos.d = list(d1 = pred$pos.d1,
                                           d2 = pred$pos.d2,
                                           d3 = pred$pos.d3)),
              pred.theta = list(qtheta = list(th1 = pred$qtheta.th1,
                                              th2 = pred$qtheta.th2,
                                              th3 = pred$qtheta.th3),
                                pos.t = list(th1 = pred$pos.th1,
                                             th2 = pred$pos.th2,
                                             th3 = pred$pos.th3)),
              dyn_scores = pred$dyn_scores))
}

.predictability <- function(dim, theta, out_len) {
  pos <- c(1:length(dim))

  # dim
  pos.d <- .terciles_position(dim)
  qdim <- list(dim[pos.d[[1]]], dim[pos.d[[2]]], dim[pos.d[[3]]])

  # theta
  pos.t <- .terciles_position(theta)
  qtheta <- list(theta[pos.t[[1]]], theta[pos.t[[2]]], theta[pos.t[[3]]])
  
  # scores position and values
  dyn_scores <- c(1:length(dim))
  values <- c(9/9, 8/9, 6/9, 7/9, 4/9, 3/9, 5/9, 2/9, 1/9)
  k = 0
  for (i in 1:3) {
    for (j in 1:3) {
      dth <- pos.d[[i]][which(pos.d[[i]] %in% pos.t[[j]])]
      k = k + 1
      dyn_scores[which(pos %in% dth)] <- values[k]
    }
    pos.d[[i]] <- c(pos.d[[i]], rep(NA, out_len[1] - length(pos.d[[i]])))
    pos.t[[i]] <- c(pos.t[[i]], rep(NA, out_len[2] - length(pos.t[[i]])))
  }

return(list(qdim.d1 = dim[pos.d[[1]]], qdim.d2 = dim[pos.d[[2]]], 
            qdim.d3 = dim[pos.d[[3]]], pos.d1 = pos.d[[1]], pos.d2 = pos.d[[1]], 
            pos.d3 = pos.d[[3]], qtheta.th1 = theta[pos.t[[1]]], 
            qtheta.th2 = theta[pos.t[[2]]], qtheta.th3 = theta[pos.t[[3]]], 
            pos.th1 = pos.t[[1]], pos.th2 = pos.t[[2]], pos.th3 = pos.t[[3]],
            dyn_scores = dyn_scores))
}

.terciles_position <- function(x) {
  q1 <- quantile(x, 0.33, na.rm = TRUE)
  q3 <- quantile(x, 0.66, na.rm = TRUE)
  t3 <- which(x >= q3)
  t1 <- which(x <= q1)
  t2 <- which(x > q1 & x < q3)
return(list(t1, t2, t3))
}
